package com.jpattern.orm.persistor;

import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jpattern.orm.exception.OrmConfigurationException;
import com.jpattern.orm.mapper.IOrmClassToolMap;
import com.jpattern.orm.mapper.clazz.ClassField;
import com.jpattern.orm.mapper.clazz.IClassField;
import com.jpattern.orm.mapper.clazz.IClassMap;
import com.jpattern.orm.persistor.generator.GeneratorManipulator;
import com.jpattern.orm.persistor.generator.GeneratorManipulatorImpl;
import com.jpattern.orm.persistor.generator.NullGeneratorManipulator;
import com.jpattern.orm.persistor.type.TypeFactory;
import com.jpattern.orm.persistor.type.TypeWrapperJdbcReady;
import com.jpattern.orm.persistor.version.NullVersionManipulator;
import com.jpattern.orm.persistor.version.VersionManipulator;
import com.jpattern.orm.persistor.version.VersionManipulatorImpl;
import com.jpattern.orm.persistor.version.VersionMath;
import com.jpattern.orm.persistor.version.VersionMathFactory;

/**
 * 
 * @author Francesco Cina'
 *
 * Mar 24, 2012
 */
public class PersistorGeneratorImpl<BEAN> implements PersistorGenerator<BEAN> {

    private final Logger logger = LoggerFactory.getLogger(getClass());

    private final IClassMap<BEAN> classMap;
    private final TypeFactory typeFactory;
    private IOrmClassToolMap ormClassToolMap;

    public PersistorGeneratorImpl(final IOrmClassToolMap ormClassToolMap, final IClassMap<BEAN> classMap, final TypeFactory typeFactory) {
        this.ormClassToolMap = ormClassToolMap;
        this.classMap = classMap;
        this.typeFactory = typeFactory;
    }

    @Override
    public OrmPersistor<BEAN> generate() throws Exception {
        Map<String, PropertyPersistor<BEAN, ?, ?>> propertyPersistorMap = this.buildPropertyPersistorMap();
        return new OrmPersistorImpl<BEAN>(this.classMap, propertyPersistorMap, this.buildVersionManipulator(propertyPersistorMap), buildGeneratorManipulator(propertyPersistorMap));
    }

    private <P, DB> Map<String, PropertyPersistor<BEAN, ?, ? >> buildPropertyPersistorMap() throws SecurityException, IllegalArgumentException {
        Map<String, PropertyPersistor<BEAN, ?, ?>> propertyPersistors = new HashMap<String, PropertyPersistor<BEAN,?,?>>();
        for (final String columnJavaName : this.classMap.getAllColumnJavaNames()) {
            final IClassField<BEAN, P> classField = this.classMap.getClassFieldByJavaName(columnJavaName);
            if (classField.getRelationVersusClass()!=null) {
                logger.debug("Build PropertyPersistor for field [{}] that is a relation versus [{}]", classField.getFieldName(), classField.getRelationVersusClass()); //$NON-NLS-1$
                Class<P> versusClass = classField.getRelationVersusClass();
                IClassMap<P> versusClassMap = ormClassToolMap.getOrmClassTool(versusClass).getClassMap();
                ClassField<P, Object> versusPkField = versusClassMap.getClassFieldByJavaName(versusClassMap.getPrimaryKeyColumnJavaNames()[0]);
                PropertyPersistor<P, Object, DB> decoratedPropertyPersistor = getPropertyPersistor(versusPkField);
                propertyPersistors.put(columnJavaName, new PropertyPeristorDecorator<P, BEAN, Object, DB>( decoratedPropertyPersistor, classField) );

            } else {
                propertyPersistors.put(columnJavaName, getPropertyPersistor(classField) );
            }
        }
        return propertyPersistors;
    }

    private VersionManipulator<BEAN> buildVersionManipulator(final Map<String, PropertyPersistor<BEAN, ?, ? >> propertyPersistorMap) {
        VersionManipulator<BEAN> versionManipulator = new NullVersionManipulator<BEAN>();

        for (final String columnJavaName : this.classMap.getAllColumnJavaNames()) {
            final IClassField<BEAN, ?> classField = this.classMap.getClassFieldByJavaName(columnJavaName);
            if (classField.getVersionInfo().isVersionable()) {
                versionManipulator = new VersionManipulatorImpl<BEAN>(propertyPersistorMap.get(classField.getFieldName()));
                break;
            }
        }

        return versionManipulator;
    }

    private <P> GeneratorManipulator<BEAN> buildGeneratorManipulator(final Map<String, PropertyPersistor<BEAN, ?, ?>> propertyPersistors) throws OrmConfigurationException, SecurityException {
        if (this.classMap.getAllGeneratedColumnJavaNames().length>0) {
            final String columnJavaName = this.classMap.getAllGeneratedColumnJavaNames()[0];
            final PropertyPersistor<BEAN, P, ?> fieldManipulator = (PropertyPersistor<BEAN, P, ?>) propertyPersistors.get(columnJavaName);
            return new GeneratorManipulatorImpl<BEAN,P>(fieldManipulator);
        }
        return new NullGeneratorManipulator<BEAN>();
    }

    private <THIS_BEAN, P, DB> PropertyPersistor<THIS_BEAN, P, DB> getPropertyPersistor(final IClassField<THIS_BEAN, P> classField){
        logger.debug("Build PropertyPersistor for field [{}]", classField.getFieldName()); //$NON-NLS-1$
        VersionMath<P> versionMath = new VersionMathFactory().getMath(classField.getType(), classField.getVersionInfo().isVersionable());
        TypeWrapperJdbcReady<P, DB> typeWrapper = this.typeFactory.getTypeWrapper(classField.getType());
        return new PropertyPersistorImpl<THIS_BEAN, P, DB>(typeWrapper, classField, versionMath) ;

    }
}
