package com.jpattern.orm.query.delete;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jpattern.orm.mapper.IOrmClassTool;
import com.jpattern.orm.mapper.IOrmClassToolMap;
import com.jpattern.orm.mapper.clazz.ClassField;
import com.jpattern.orm.mapper.clazz.IClassMap;
import com.jpattern.orm.mapper.relation.FKClassField;
import com.jpattern.orm.mapper.relation.RelationOuterFK;
import com.jpattern.orm.query.NameSolver;
import com.jpattern.orm.query.OrmClassToolMapNameSolver;
import com.jpattern.orm.query.SmartRenderableSqlQuery;
import com.jpattern.orm.query.clause.where.ExpressionElement;
import com.jpattern.orm.query.find.CustomFindQuery;
import com.jpattern.orm.session.ResultSetRowReader;
import com.jpattern.orm.session.SessionSqlPerformer;
import com.jpattern.orm.session.SqlPerformer;

/**
 * 
 * @author Francesco Cina
 *
 * 10/lug/2011
 */
public class DeleteQueryOrm<BEAN> extends SmartRenderableSqlQuery implements DeleteQuery<BEAN> {

    private Logger logger = LoggerFactory.getLogger(getClass());
    private boolean _cascade = false;
    private final DeleteWhereImpl<BEAN> where = new DeleteWhereImpl<BEAN>(this);
    private final IOrmClassToolMap ormClassToolMap;
    private final SessionSqlPerformer session;
    private int _queryTimeout = 0;
    private final Class<?> clazz;
    private final NameSolver nameSolver;
    private IOrmClassTool<BEAN> ormClassTool;

    public DeleteQueryOrm(final Class<BEAN> clazz, final IOrmClassToolMap ormClassToolMap, final SessionSqlPerformer session) {
        this.clazz = clazz;
        this.ormClassToolMap = ormClassToolMap;
        this.session = session;
        nameSolver = new OrmClassToolMapNameSolver(ormClassToolMap, true);
        nameSolver.register(clazz, clazz.getSimpleName() + ALIAS_FIXED_SUFFIX);
        ormClassTool = ormClassToolMap.getOrmClassTool(clazz);
    }

    @Override
    public DeleteWhere<BEAN> where() {
        return where;
    }

    @Override
    public int now() {
        int deleted = deleteOuterRelations();
        Map<String, Object> innerRelationValues = fetchInnerRelationValue();

        final List<Object> values = new ArrayList<Object>();
        appendValues(values);
        final SqlPerformer sqlExec = session.sqlPerformer();
        sqlExec.setQueryTimeout(getQueryTimeout());
        deleted+= sqlExec.update(renderSql(), values);

        deleted+= deleteInnerRelations(innerRelationValues);
        return deleted;
    }

    private int deleteOuterRelations() {
        int deleted = 0;
        if (_cascade) {
            List<RelationOuterFK<BEAN, ?, ?>> relations = ormClassTool.getClassMap().getOuterRelations();

            for (RelationOuterFK<BEAN, ?, ?> relation : relations) {
                logger.debug("Delete related beans of class [{}]", relation.getRelationWithClass()); //$NON-NLS-1$
                Class<?> relationWith = relation.getRelationWithClass();
                String relationFieldName = ormClassToolMap.getOrmClassTool(relationWith).getClassMap().getFKs().versus(clazz).getBeanFieldName();

                String alias = clazz.getSimpleName() + "_1"; //$NON-NLS-1$
                final String pkName = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames()[0];
                CustomFindQuery idQuery = session.findQuery(pkName, clazz, alias);
                for (ExpressionElement expressionElement : where.getElementList()) {
                    idQuery.where().and(expressionElement);
                }
                //                List<Object> ids = idQuery.find(new ResultSetRowReader<Object>() {
                //                    @Override
                //                    public Object readRow(final ResultSet rs, final int rowNum) throws SQLException {
                //                        return rs.getObject(pkName);
                //                    }
                //                });
                deleted+=session.deleteQuery(relationWith).cascade(_cascade).where().in(relationFieldName, idQuery).now();
            }
        }
        return deleted;
    }


    private Map<String, Object> fetchInnerRelationValue() {
        final Map<String, Object> values = new HashMap<String, Object>();
        List<FKClassField<BEAN, ?>> relations = ormClassTool.getClassMap().getInnerRelations();
        if (_cascade && !relations.isEmpty()) {
            final String[] relationFieldNames = new String[relations.size()];
            for (int i=0; i<relations.size(); i++) {
                relationFieldNames[i] = relations.get(i).getFieldName();
            }
            String alias = clazz.getSimpleName() + "_1"; //$NON-NLS-1$
            ResultSetRowReader<Void> rsrr = new ResultSetRowReader<Void>() {
                @Override
                public Void readRow(final ResultSet rs, final int rowNum) throws SQLException {
                    for (String relationFieldName : relationFieldNames) {
                        values.put(relationFieldName, rs.getObject(relationFieldName) );
                    }
                    return null;
                }
            };
            CustomFindQuery find = session.findQuery(relationFieldNames, clazz, alias);
            for (ExpressionElement expressionElement : where.getElementList()) {
                find.where().and(expressionElement);
            }
            find.get(rsrr);


            if (logger.isDebugEnabled()) {
                for (String relationFieldName : relationFieldNames) {
                    logger.debug("Loaded value for inner FKs field [{}] , value [{}]", relationFieldName, values.get(relationFieldName)); //$NON-NLS-1$
                }
            }
        }
        return values;
    }

    /**
     * @return
     */
    private <RELATION_WITH> int deleteInnerRelations(final Map<String, Object> innerRelationValues) {
        int deleted = 0;
        if (_cascade) {
            for (Entry<String, Object> innerRelationEntry : innerRelationValues.entrySet()) {
                Object fieldValue = innerRelationEntry.getValue();
                if (fieldValue!=null) {
                    String fieldName = innerRelationEntry.getKey();
                    ClassField<BEAN, Object> field = ormClassTool.getClassMap().getClassFieldByJavaName(fieldName);
                    Class<RELATION_WITH> relationWith = field.getRelationVersusClass();
                    IClassMap<RELATION_WITH> relationWithClassMap = ormClassToolMap.getOrmClassTool(relationWith).getClassMap();
                    String relationWithPkFieldName = relationWithClassMap.getPrimaryKeyColumnJavaNames()[0];
                    deleted+=session.deleteQuery(relationWith).cascade(_cascade).where().eq(relationWithPkFieldName, fieldValue).now();
                }
            }
        }
        return deleted;
    }

    @Override
    public final void appendValues(final List<Object> values) {
        where.appendElementValues(values);
    }

    @Override
    public DeleteQuery<BEAN> queryTimeout(final int queryTimeout) {
        this._queryTimeout = queryTimeout;
        return this;
    }

    public int getQueryTimeout() {
        return _queryTimeout;
    }

    @Override
    public int getStatusVersion() {
        return where.getElementStatusVersion();
    }

    @Override
    public final void doRender(final StringBuilder queryBuilder) {
        queryBuilder.append("DELETE FROM "); //$NON-NLS-1$
        queryBuilder.append(ormClassToolMap.getOrmClassTool(clazz).getClassMap().getTableInfo().getTableNameWithSchema() );
        queryBuilder.append(" "); //$NON-NLS-1$
        where.renderSqlElement(queryBuilder, nameSolver);
    }

    @Override
    public DeleteQuery<BEAN> cascade(final boolean cascade) {
        _cascade = cascade;
        return this;
    }

}
