package com.jpattern.orm.query.find;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.jpattern.orm.cache.CacheManager;
import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.exception.OrmNotUniqueResultException;
import com.jpattern.orm.exception.OrmNotUniqueResultManyResultsException;
import com.jpattern.orm.exception.OrmNotUniqueResultNoResultException;
import com.jpattern.orm.mapper.IOrmClassTool;
import com.jpattern.orm.mapper.IOrmClassToolMap;
import com.jpattern.orm.mapper.clazz.ClassField;
import com.jpattern.orm.mapper.relation.RelationOuterFK;
import com.jpattern.orm.persistor.BeanFromResultSet;
import com.jpattern.orm.persistor.reflection.SetManipulator;
import com.jpattern.orm.query.LockMode;
import com.jpattern.orm.query.NameSolver;
import com.jpattern.orm.query.OrmClassToolMapNameSolver;
import com.jpattern.orm.query.OrmRowMapper;
import com.jpattern.orm.query.SmartRenderableSqlQuery;
import com.jpattern.orm.query.find.cache.CacheStrategy;
import com.jpattern.orm.query.find.cache.CacheStrategyCallback;
import com.jpattern.orm.query.find.cache.CacheStrategyEntry;
import com.jpattern.orm.query.find.cache.CacheStrategyImpl;
import com.jpattern.orm.query.find.cache.NullCacheStrategy;
import com.jpattern.orm.session.ResultSetReader;
import com.jpattern.orm.session.SessionSqlPerformer;
import com.jpattern.orm.session.SqlPerformer;
import com.jpattern.orm.util.GenericWrapper;

/**
 * 
 * @author Francesco Cina
 *
 * 20/giu/2011
 */
public class FindQueryOrm<BEAN> extends SmartRenderableSqlQuery implements FindQuery<BEAN> {

    private boolean _lazy = true;
    private final IOrmClassToolMap ormClassToolMap;
    private final Class<BEAN> clazz;
    private final SessionSqlPerformer session;
    private int _queryTimeout = 0;
    private int mainMaxRows = 0;
    private LockMode _lockMode = LockMode.NO_LOCK;
    private final CustomFindSelectImpl select;
    private final FindWhereImpl<BEAN> where = new FindWhereImpl<BEAN>(this);
    private final FindOrderByImpl<BEAN> orderBy = new FindOrderByImpl<BEAN>(this);
    private final FindFromImpl<BEAN> from;
    private int versionStatus = 0;
    private final NameSolver nameSolver;
    private final CacheManager cacheManager;
    private CacheStrategy<BEAN> cacheStrategy = new NullCacheStrategy<BEAN>();
    private List<String> _ignoredFields = new ArrayList<String>();

    public FindQueryOrm(final IOrmClassToolMap ormClassToolMap, final SessionSqlPerformer session, final Class<BEAN> clazz, final CacheManager cacheManager, final String alias) {
        this.ormClassToolMap = ormClassToolMap;
        this.session = session;
        this.clazz = clazz;
        this.cacheManager = cacheManager;

        this.nameSolver = new OrmClassToolMapNameSolver(ormClassToolMap, false);
        this.from = new FindFromImpl<BEAN>(this, ormClassToolMap, clazz, nameSolver.register(clazz, alias), nameSolver);
        this.select = new CustomFindSelectImpl(ormClassToolMap.getOrmClassTool(clazz).getClassMap().getAllColumnJavaNames());

    }

    @Override
    public final void appendValues(final List<Object> values) {
        this.where.appendElementValues(values);
    }

    @Override
    public final void doRender(final StringBuilder queryBuilder) {
        this.select.ignore(_ignoredFields);
        this.select.renderSqlElement(queryBuilder, nameSolver);
        this.from.renderSqlElement(queryBuilder, nameSolver);
        this.where.renderSqlElement(queryBuilder, nameSolver);
        this.orderBy.renderSqlElement(queryBuilder,nameSolver);
        queryBuilder.append(this._lockMode.getMode());
    }

    @Override
    public BEAN get() throws OrmException, OrmNotUniqueResultException {
        final GenericWrapper<BEAN> wrapper = new GenericWrapper<BEAN>(null);
        OrmRowMapper<BEAN> srr = new OrmRowMapper<BEAN>() {
            @Override
            public void read(final BEAN newObject, final int rowCount) {
                wrapper.setValue(newObject);
            }
        };
        find(srr, 1);
        return wrapper.getValue();
    }

    @Override
    public void get(final OrmRowMapper<BEAN> srr) throws OrmException {
        find(srr, this.getMaxRows());
    }

    public void find(final OrmRowMapper<BEAN> srr, final int maxRows) throws OrmException {
        final List<Object> values = new ArrayList<Object>();
        this.where.appendElementValues(values);
        final String sql = this.renderSql();

        cacheStrategy.find(sql, values, _ignoredFields, maxRows, srr, new CacheStrategyCallback<BEAN>() {

            @Override
            public void doWhenNotInCache(final CacheStrategyEntry<BEAN> cacheStrategyEntry) {
                final ResultSetReader<Object> resultSetReader = new ResultSetReader<Object>() {
                    @Override
                    public Object read(final ResultSet resultSet) throws SQLException {
                        int rowCount = 0;
                        final IOrmClassTool<BEAN> ormClassTool = ormClassToolMap.getOrmClassTool(clazz);
                        while ( resultSet.next() ) {
                            BeanFromResultSet<BEAN> beanFromRS = ormClassTool.getOrmPersistor().beanFromResultSet(resultSet, _ignoredFields);
                            loadInnerRelations( beanFromRS.getBean(), beanFromRS.getInnerFkValues(), ormClassTool );
                            loadOuterRelations(beanFromRS.getBean(), ormClassTool);
                            srr.read( beanFromRS.getBean() , rowCount );
                            rowCount++;
                        }
                        return null;
                    }

                };

                final SqlPerformer sqlExec = session.sqlPerformer();
                sqlExec.setMaxRows(maxRows);
                sqlExec.setQueryTimeout(getQueryTimeout());
                sqlExec.query(sql, resultSetReader, values);
            }

        });


    }

    private void loadOuterRelations(final BEAN bean, final IOrmClassTool<BEAN> ormClassTool) {
        if (!_lazy) {
            List<RelationOuterFK<BEAN, ?, ?>> relations = ormClassTool.getClassMap().getOuterRelations();
            if (!relations.isEmpty()) {
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                Object beanPrimaryKey = ormClassTool.getOrmPersistor().getPropertyValues(pks, bean)[0];
                for (RelationOuterFK<BEAN, ?, ?> relation : relations) {
                    Class<?> relationWith = relation.getRelationWithClass();
                    String relationFieldName = relation.getRelationClassField().getFieldName();

                    FindWhere<?> query = session.findQuery(relationWith).lazy(_lazy).where().eq(relationFieldName, beanPrimaryKey);
                    Object relatedBeans = null;
                    if (relation.isOneToMany()) {
                        relatedBeans = query.getList();
                    } else {
                        relatedBeans = query.get();
                    }
                    (( SetManipulator<BEAN, Object>) relation.getSetManipulator()).setValue(bean, relatedBeans);
                }
            }
        }
    }

    private <RELATION_VERSUS> void loadInnerRelations(final BEAN bean, final Map<String, Object> innerFkValues, final IOrmClassTool<BEAN> ormClassTool) {
        if (!_lazy) {
            for (Entry<String, Object> innerFkEntry : innerFkValues.entrySet()) {
                Object innerFkValue = innerFkEntry.getValue();
                if (innerFkValue!=null) {
                    String innerFkKey = innerFkEntry.getKey();
                    ClassField<BEAN, RELATION_VERSUS> innerFkField = ormClassTool.getClassMap().getClassFieldByJavaName(innerFkKey);
                    Class<RELATION_VERSUS> relationWith = innerFkField.getRelationVersusClass();
                    innerFkField.getSetManipulator().setValue(bean, session.find(relationWith, innerFkValue).get());
                }
            }
        }
    }

    @Override
    public List<BEAN> getList() {
        final List<BEAN> results = new ArrayList<BEAN>();
        OrmRowMapper<BEAN> srr = new OrmRowMapper<BEAN>() {
            @Override
            public void read(final BEAN newObject, final int rowCount) {
                results.add(newObject);
            }
        };
        get(srr);
        return results;
    }

    @Override
    public int getRowCount() {
        final List<Object> values = new ArrayList<Object>();
        this.where.appendElementValues(values);
        final SqlPerformer sqlExec = this.session.sqlPerformer();
        sqlExec.setMaxRows(this.getMaxRows());
        sqlExec.setQueryTimeout(this.getQueryTimeout());
        return sqlExec.queryForInt(this.renderRowCountSql(), values);
    }

    @Override
    public BEAN getUnique() throws OrmNotUniqueResultException {
        final GenericWrapper<BEAN> wrapper = new GenericWrapper<BEAN>(null);
        final GenericWrapper<Integer> resultsCount = new GenericWrapper<Integer>(-1);
        OrmRowMapper<BEAN> srr = new OrmRowMapper<BEAN>() {
            @Override
            public void read(final BEAN newObject, final int rowCount) {
                resultsCount.setValue(rowCount);
                wrapper.setValue(newObject);
            }
        };
        find(srr, 2);
        if ( resultsCount.getValue() != 0) {
            if ( resultsCount.getValue() == -1 ) {
                throw new OrmNotUniqueResultNoResultException("The query execution returned a number of rows different than one: no results found"); //$NON-NLS-1$
            }
            throw new OrmNotUniqueResultManyResultsException("The query execution returned a number of rows different than one: more than one result found"); //$NON-NLS-1$
        }
        return wrapper.getValue();
    }

    @Override
    public FindQuery<BEAN> fullOuterJoin(final Class<?> joinClass) {
        return this.from.fullOuterJoin(joinClass);
    }

    @Override
    public FindQuery<BEAN> fullOuterJoin(final Class<?> joinClass,
            final String joinClassAlias) {
        return this.from.fullOuterJoin(joinClass, joinClassAlias);
    }

    @Override
    public FindQuery<BEAN> fullOuterJoin(final Class<?> joinClass,
            final String onLeftProperty, final String onRigthProperty) {
        return this.from.fullOuterJoin(joinClass, onLeftProperty, onRigthProperty);
    }

    @Override
    public FindQuery<BEAN> fullOuterJoin(final Class<?> joinClass,
            final String joinClassAlias, final String onLeftProperty, final String onRigthProperty) {
        return this.from.fullOuterJoin(joinClass, joinClassAlias, onLeftProperty,
                onRigthProperty);
    }

    @Override
    public String renderRowCountSql() {
        final StringBuilder queryBuilder = new StringBuilder();
        queryBuilder.append("SELECT COUNT(*) "); //$NON-NLS-1$
        this.from.renderSqlElement(queryBuilder, nameSolver);
        this.where.renderSqlElement(queryBuilder, nameSolver);
        return queryBuilder.toString();
    }

    public LockMode getLockMode() {
        return this._lockMode;
    }

    public final int getMaxRows() throws OrmException {
        return this.mainMaxRows;
    }

    public final int getQueryTimeout() {
        return this._queryTimeout;
    }

    @Override
    public final int getStatusVersion() {
        return this.versionStatus + select.getElementStatusVersion() + this.from.getElementStatusVersion() + this.where.getElementStatusVersion() + this.orderBy.getElementStatusVersion();

    }

    @Override
    public FindQuery<BEAN> innerJoin(final Class<?> joinClass) {
        return this.from.innerJoin(joinClass);
    }

    @Override
    public FindQuery<BEAN> innerJoin(final Class<?> joinClass, final String joinClassAlias) {
        return this.from.innerJoin(joinClass, joinClassAlias);
    }

    @Override
    public FindQuery<BEAN> innerJoin(final Class<?> joinClass, final String onLeftProperty,
            final String onRigthProperty) {
        return this.from.innerJoin(joinClass, onLeftProperty, onRigthProperty);
    }

    @Override
    public FindQuery<BEAN> innerJoin(final Class<?> joinClass, final String joinClassAlias,
            final String onLeftProperty, final String onRigthProperty) {
        return this.from.innerJoin(joinClass, joinClassAlias, onLeftProperty,
                onRigthProperty);
    }

    public boolean isDistinct() throws OrmException {
        return select.isDistinct();
    }

    @Override
    public FindQuery<BEAN> join(final Class<?> joinClass) {
        return this.from.join(joinClass);
    }

    @Override
    public FindQuery<BEAN> join(final Class<?> joinClass, final String joinClassAlias) {
        return this.from.join(joinClass, joinClassAlias);
    }

    @Override
    public FindQuery<BEAN> leftOuterJoin(final Class<?> joinClass) {
        return this.from.leftOuterJoin(joinClass);
    }

    @Override
    public FindQuery<BEAN> leftOuterJoin(final Class<?> joinClass,
            final String joinClassAlias) {
        return this.from.leftOuterJoin(joinClass, joinClassAlias);
    }

    @Override
    public FindQuery<BEAN> leftOuterJoin(final Class<?> joinClass,
            final String onLeftProperty, final String onRigthProperty) {
        return this.from.leftOuterJoin(joinClass, onLeftProperty, onRigthProperty);
    }

    @Override
    public FindQuery<BEAN> leftOuterJoin(final Class<?> joinClass,
            final String joinClassAlias, final String onLeftProperty, final String onRigthProperty) {
        return this.from.leftOuterJoin(joinClass, joinClassAlias, onLeftProperty,
                onRigthProperty);
    }

    @Override
    public FindQuery<BEAN> naturalJoin(final Class<?> joinClass) {
        return this.from.naturalJoin(joinClass);
    }

    @Override
    public FindQuery<BEAN> naturalJoin(final Class<?> joinClass, final String joinClassAlias) {
        return this.from.naturalJoin(joinClass, joinClassAlias);
    }

    @Override
    public final FindOrderBy<BEAN> orderBy() throws OrmException {
        return this.orderBy;
    }

    @Override
    public FindQuery<BEAN> rightOuterJoin(final Class<?> joinClass) {
        return this.from.rightOuterJoin(joinClass);
    }

    @Override
    public FindQuery<BEAN> rightOuterJoin(final Class<?> joinClass,
            final String joinClassAlias) {
        return this.from.rightOuterJoin(joinClass, joinClassAlias);
    }

    @Override
    public FindQuery<BEAN> rightOuterJoin(final Class<?> joinClass,
            final String onLeftProperty, final String onRigthProperty) {
        return this.from.rightOuterJoin(joinClass, onLeftProperty, onRigthProperty);
    }

    @Override
    public FindQuery<BEAN> rightOuterJoin(final Class<?> joinClass,
            final String joinClassAlias, final String onLeftProperty, final String onRigthProperty) {
        return this.from.rightOuterJoin(joinClass, joinClassAlias, onLeftProperty,
                onRigthProperty);
    }

    @Override
    public FindQuery<BEAN> distinct(final boolean distinct) {
        select.setDistinct(distinct);
        return this;
    }

    @Override
    public FindQuery<BEAN> lockMode(final LockMode lockMode) {
        this._lockMode = lockMode;
        this.versionStatus++;
        return this;
    }

    @Override
    public final FindQuery<BEAN> maxRows(final int maxRows) throws OrmException {
        this.mainMaxRows = maxRows;
        return this;
    }

    @Override
    public final FindQuery<BEAN> queryTimeout(final int queryTimeout) {
        this._queryTimeout = queryTimeout;
        return this;
    }

    @Override
    public final FindWhere<BEAN> where() throws OrmException {
        return this.where;
    }

    @Override
    public FindQuery<BEAN> lazy(final boolean lazy) {
        this._lazy = lazy;
        return this;
    }

    @Override
    public FindQuery<BEAN> cache(final String cache) {
        if ((cache!=null) && !cache.isEmpty()) {
            cacheStrategy = new CacheStrategyImpl<BEAN>(cacheManager.getCache(cache));
        }
        return this;
    }

    @Override
    public final FindQuery<BEAN> ignore(final String... fields) {
        return ignore(true, fields);
    }

    @Override
    public final FindQuery<BEAN> ignore(final boolean ignoreFieldsCondition, final String... fields) {
        if(ignoreFieldsCondition) {
            _ignoredFields = Arrays.asList(fields);
        }
        return this;
    }

}
