package com.jpattern.orm.query.save;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.jpattern.orm.dialect.DBProfile;
import com.jpattern.orm.exception.OrmReflectionException;
import com.jpattern.orm.mapper.IOrmClassTool;
import com.jpattern.orm.mapper.IOrmClassToolMap;
import com.jpattern.orm.mapper.clazz.IClassField;
import com.jpattern.orm.mapper.clazz.IClassMap;
import com.jpattern.orm.mapper.relation.FKClassField;
import com.jpattern.orm.mapper.relation.RelationOuterFK;
import com.jpattern.orm.persistor.OrmPersistor;
import com.jpattern.orm.persistor.reflection.SetManipulator;
import com.jpattern.orm.session.GeneratedKeyReader;
import com.jpattern.orm.session.SessionSqlPerformer;
import com.jpattern.orm.session.SqlPerformer;

/**
 * 
 * @author Francesco Cina
 *
 * 10/lug/2011
 */
public class SaveQueryOrm<BEAN> implements SaveQuery<BEAN> {

    private final SessionSqlPerformer session;
    private int queryTimeout = 0;
    private final Class<BEAN> clazz;
    private final IOrmClassTool<BEAN> ormClassTool;
    private final BEAN bean;
    private final DBProfile dbProfile;
    private boolean useGenerator = false;
    private boolean _cascade = false;
    private boolean _saveOrUpdate = false;

    public SaveQueryOrm(final BEAN bean, final IOrmClassToolMap ormClassToolMap, final SessionSqlPerformer session, final DBProfile dbProfile) {
        this.bean = bean;
        this.dbProfile = dbProfile;
        this.clazz = (Class<BEAN>) bean.getClass();
        this.ormClassTool = ormClassToolMap.getOrmClassTool(clazz);
        this.session = session;
    }

    @Override
    public BEAN perform() {
        saveInnerRelations();

        final OrmPersistor<BEAN> persistor = ormClassTool.getOrmPersistor();
        final SqlPerformer sqlExec = this.session.sqlPerformer();
        sqlExec.setQueryTimeout(queryTimeout);

        //CHECK IF OBJECT HAS A 'VERSION' FIELD and increase it
        persistor.increaseVersion(bean, true);
        String sql = renderSql();
        if (!useGenerator) {
            String[] keys = ormClassTool.getClassMap().getAllColumnJavaNames();
            Object[] values = persistor.getPropertyValues(keys, bean);
            sqlExec.update(sql, values);
        } else {
            final GeneratedKeyReader generatedKeyExtractor = new GeneratedKeyReader() {

                @Override
                public void read(final ResultSet generatedKeyResultSet) throws SQLException {
                    if (generatedKeyResultSet.next()) {
                        persistor.updateGeneratedValues(generatedKeyResultSet, bean);
                    }
                }

                @Override
                public String[] generatedColumnNames() {
                    return ormClassTool.getClassMap().getAllGeneratedColumnDBNames();
                }
            };
            String[] keys = ormClassTool.getClassMap().getAllNotGeneratedColumnJavaNames();
            Object[] values = persistor.getPropertyValues(keys, bean);
            sqlExec.update(sql, generatedKeyExtractor, values);
        }
        saveOuterRelations();
        return bean;

    }

    @SuppressWarnings("unchecked")
    private <RELATION> void saveInnerRelations() {
        if (_cascade) {
            List<FKClassField<BEAN, ?>> relations = ormClassTool.getClassMap().getInnerRelations();
            for (FKClassField<BEAN, ?> fkClassField : relations) {
                FKClassField<BEAN, RELATION> relationField = (FKClassField<BEAN, RELATION>) fkClassField;
                RELATION innerBean = relationField.getGetManipulator().getValue(bean);
                RELATION saved;
                if (_saveOrUpdate) {
                    saved = session.saveOrUpdate(innerBean).cascade(_cascade).now();
                } else {
                    saved = session.save(innerBean).cascade(_cascade).now();
                }
                relationField.getSetManipulator().setValue(bean, saved);
            }
        }
    }

    private <RELATION> void saveOuterRelations() {
        if (_cascade) {
            List<RelationOuterFK<BEAN, ?, ?>> relations = ormClassTool.getClassMap().getOuterRelations();
            if (!relations.isEmpty()) {
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                Object beanPrimaryKeyValue = ormClassTool.getOrmPersistor().getPropertyValues(pks, bean)[0];
                for (RelationOuterFK<BEAN,? ,?> relation : relations) {

                    IClassField<RELATION, Object> relationClassField = (IClassField<RELATION, Object>) relation.getRelationClassField();
                    Object saved = null;
                    if (relation.isOneToMany()) {
                        Collection<RELATION> relatedBeans = (Collection<RELATION>) relation.getGetManipulator().getValue(bean);
                        for (RELATION relatedBean : relatedBeans) {
                            relationClassField.getSetManipulator().setValue(relatedBean, beanPrimaryKeyValue);
                        }
                        if (_saveOrUpdate) {
                            saved = session.saveOrUpdate(relatedBeans).cascade(_cascade).now();
                        } else {
                            saved = session.save(relatedBeans).cascade(_cascade).now();
                        }
                    } else {
                        RELATION relatedBean = (RELATION) relation.getGetManipulator().getValue(bean);
                        if (relatedBean!=null) {
                            relationClassField.getSetManipulator().setValue(relatedBean, beanPrimaryKeyValue);
                            if (_saveOrUpdate) {
                                saved = session.saveOrUpdate(relatedBean).cascade(_cascade).now();
                            } else {
                                saved = session.save(relatedBean).cascade(_cascade).now();
                            }
                        }
                    }
                    (( SetManipulator<BEAN, Object>) relation.getSetManipulator()).setValue(bean, saved);
                }
            }
        }
    }

    @Override
    public SaveQuery<BEAN> setQueryTimeout(final int queryTimeout) {
        this.queryTimeout = queryTimeout;
        return this;
    }

    @Override
    public int getQueryTimeout() {
        return this.queryTimeout;
    }

    @Override
    public SaveQuery<BEAN> value(final String propertyName, final Object propertyValue) {
        try {
            ormClassTool.getClassMap().getClassFieldByJavaName(propertyName).getSetManipulator().setValue(bean, propertyValue);
            return this;
        } catch (Exception e) {
            throw new OrmReflectionException(e);
        }
    }

    @Override
    public SaveQuery<BEAN> values(final Map<String, Object> properties) {
        for (Entry<String, Object> propertyEntry : properties.entrySet()) {
            value(propertyEntry.getKey(), propertyEntry.getValue());
        }
        return this;
    }

    @SuppressWarnings("nls")
    public final String renderSql() {
        updateUseGenerators();
        return generateSaveQuery(!useGenerator); //$NON-NLS-1$
    }

    private String generateSaveQuery(final boolean ignoreGenerators) {
        IClassMap<BEAN> classMap = ormClassTool.getClassMap();
        final StringBuilder builder = new StringBuilder("INSERT INTO "); //$NON-NLS-1$
        builder.append(ormClassTool.getClassMap().getTableInfo().getTableNameWithSchema());
        builder.append(" ("); //$NON-NLS-1$
        builder.append( this.columnToCommaSepareted( "" , classMap.getAllColumnJavaNames(), ignoreGenerators )); //$NON-NLS-1$
        builder.append(") VALUES ("); //$NON-NLS-1$
        builder.append( this.questionCommaSepareted( classMap.getAllColumnJavaNames(), ignoreGenerators ));
        builder.append(") "); //$NON-NLS-1$
        return builder.toString();
    }

    private void updateUseGenerators() {
        useGenerator = ormClassTool.getOrmPersistor().useGenerators(bean);
    }

    private String questionCommaSepareted(final String[] fieldNames, final boolean ignoreGenerators) {
        final StringBuilder builder = new StringBuilder();
        IClassMap<BEAN> classMap = ormClassTool.getClassMap();
        final int length = fieldNames.length;
        boolean generatedKey = false;
        if (length > 0) {
            for (int i=0; i<(length-1) ; i++) {
                IClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[i]);
                final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, dbProfile, ignoreGenerators );
                generatedKey = generatedKey || columnValueGenerator.isAutoGenerated();
                final String queryParameter = columnValueGenerator.insertQueryParameter( "?"); //$NON-NLS-1$
                if (queryParameter.length()>0) {
                    builder.append( queryParameter );
                    builder.append(", "); //$NON-NLS-1$
                }
            }
            IClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[fieldNames.length-1]);
            final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, dbProfile, ignoreGenerators );
            generatedKey = generatedKey || columnValueGenerator.isAutoGenerated();
            builder.append( columnValueGenerator.insertQueryParameter( "?") ); //$NON-NLS-1$
        }
        return builder.toString();
    }

    private String columnToCommaSepareted(final String prefix, final String[] fieldNames, final boolean ignoreGenerators) {
        final StringBuilder builder = new StringBuilder();
        IClassMap<BEAN> classMap = ormClassTool.getClassMap();
        if (fieldNames.length > 0) {
            for (int i=0; i<(fieldNames.length-1) ; i++) {
                IClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[i]);
                final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, dbProfile, ignoreGenerators );
                final String queryParameter = columnValueGenerator.insertColumn(prefix + classField.getColumnInfo().getDBColumnName());
                if (queryParameter.length()>0) {
                    builder.append( queryParameter );
                    builder.append(", "); //$NON-NLS-1$
                }
            }
            IClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[fieldNames.length-1]);
            final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, dbProfile, ignoreGenerators );
            builder.append( columnValueGenerator.insertColumn( prefix + classField.getColumnInfo().getDBColumnName()) );
        }

        return builder.toString();
    }

    /**
     * @param cascade
     * @return
     */
    @Override
    public SaveQuery<BEAN> cascade(final boolean cascade) {
        this._cascade = cascade;
        return this;
    }

    @Override
    public SaveQuery<BEAN> saveOrUpdate(final boolean saveOrUpdate) {
        this._saveOrUpdate = saveOrUpdate;
        return this;
    }

}
