/* ----------------------------------------------------------------------------
 *     PROJECT : JPOrm
 *
 *  CREATED BY : Francesco Cina'
 *          ON : Feb 23, 2013
 * ----------------------------------------------------------------------------
 */
package com.jpattern.orm.query.update;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.jpattern.orm.exception.OrmOptimisticLockException;
import com.jpattern.orm.exception.OrmReflectionException;
import com.jpattern.orm.mapper.IOrmClassTool;
import com.jpattern.orm.mapper.IOrmClassToolMap;
import com.jpattern.orm.mapper.clazz.IClassField;
import com.jpattern.orm.mapper.relation.FKClassField;
import com.jpattern.orm.mapper.relation.RelationOuterFK;
import com.jpattern.orm.persistor.OrmPersistor;
import com.jpattern.orm.persistor.reflection.SetManipulator;
import com.jpattern.orm.query.find.FindWhere;
import com.jpattern.orm.session.SessionSqlPerformer;

/**
 * <class_description>
 * <p><b>notes</b>:
 * <p>ON : Feb 23, 2013
 *
 * @author Francesco Cina'
 * @version $Revision
 */
public class UpdateQueryOrm<BEAN> implements UpdateQuery<BEAN> {

    private boolean _cascade;
    private SessionSqlPerformer session;
    private BEAN bean;
    private IOrmClassTool<BEAN> ormClassTool;
    private int queryTimeout;
    private Class<BEAN> clazz;
    private boolean _saveOrUpdate = false;

    /**
     * @param newBean
     * @param ormClassToolMap
     * @param ormSession
     */
    public UpdateQueryOrm(final BEAN bean, final IOrmClassToolMap ormClassToolMap, final SessionSqlPerformer session) {
        this.bean = bean;
        this.clazz= (Class<BEAN>) bean.getClass();
        this.ormClassTool = ormClassToolMap.getOrmClassTool(clazz);
        this.session = session;
    }

    @Override
    public UpdateQuery<BEAN> value(final String propertyName, final Object value) {
        try {
            ormClassTool.getClassMap().getClassFieldByJavaName(propertyName).getSetManipulator().setValue(bean, value);
            return this;
        } catch (Exception e) {
            throw new OrmReflectionException(e);
        }
    }

    @Override
    public BEAN perform() {
        updateInnerRelations();

        final OrmPersistor<BEAN> persistor = ormClassTool.getOrmPersistor();

        //CHECK IF OBJECT HAS A 'VERSION' FIELD
        if (persistor.isVersionable()) {
            FindWhere<BEAN> query = (FindWhere<BEAN>) session.findQuery(bean.getClass()).where();
            String[] keys = ormClassTool.getClassMap().getPrimaryKeyAndVersionColumnJavaNames();
            Object[] values = persistor.getPropertyValues(keys, bean);
            for (int i = 0; i < keys.length; i++) {
                query.eq(keys[i], values[i]);
            }
            if ( query.getRowCount()==0 ) {
                throw new OrmOptimisticLockException("The bean of class [" + bean.getClass() + "] cannot be updated. Version in the DB is not the expected one."); //$NON-NLS-1$ //$NON-NLS-2$
            }
            ormClassTool.getOrmPersistor().increaseVersion(bean, false);
        }

        CustomUpdateQuery updateQuery = session.updateQuery(bean.getClass()).queryTimeout(queryTimeout);

        CustomUpdateWhere updateQueryWhere = updateQuery.where();
        String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
        Object[] pkValues = persistor.getPropertyValues(pks, bean);
        for (int i = 0; i < pks.length; i++) {
            updateQueryWhere.eq(pks[i], pkValues[i]);
        }

        CustomUpdateSet updateQuerySet = updateQuery.set();
        String[] notPks = ormClassTool.getClassMap().getNotPrimaryKeyColumnJavaNames();
        Object[] notPkValues = persistor.getPropertyValues(notPks, bean);
        for (int i = 0; i < notPks.length; i++) {
            updateQuerySet.eq(notPks[i], notPkValues[i]);
        }

        updateQuery.now();
        updateOuterRelations();
        return bean;
    }

    @SuppressWarnings("unchecked")
    private <RELATION> void updateInnerRelations() {
        if (_cascade) {
            List<FKClassField<BEAN, ?>> relations = ormClassTool.getClassMap().getInnerRelations();
            for (FKClassField<BEAN, ?> fkClassField : relations) {
                FKClassField<BEAN, RELATION> relationField = (FKClassField<BEAN, RELATION>) fkClassField;
                RELATION innerBean = relationField.getGetManipulator().getValue(bean);
                RELATION saved;
                if (_saveOrUpdate) {
                    saved = session.saveOrUpdate(innerBean).cascade(_cascade).now();
                } else {
                    saved = session.update(innerBean).cascade(_cascade).now();
                }
                relationField.getSetManipulator().setValue(bean, saved);
            }
        }
    }

    private <RELATION> void updateOuterRelations() {
        if (_cascade) {
            List<RelationOuterFK<BEAN, ?, ?>> relations = ormClassTool.getClassMap().getOuterRelations();
            if (!relations.isEmpty()) {
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                Object beanPrimaryKeyValue = ormClassTool.getOrmPersistor().getPropertyValues(pks, bean)[0];
                for (RelationOuterFK<BEAN, ?, ?> relation : relations) {

                    IClassField<RELATION, Object> relationClassField = (IClassField<RELATION, Object>) relation.getRelationClassField();

                    Object saved = null;
                    if (relation.isOneToMany()) {
                        Collection<RELATION> relatedBeans = (Collection<RELATION>) relation.getGetManipulator().getValue(bean);
                        for (RELATION relatedBean : relatedBeans) {
                            relationClassField.getSetManipulator().setValue(relatedBean, beanPrimaryKeyValue);
                        }
                        if (_saveOrUpdate) {
                            saved = session.saveOrUpdate(relatedBeans).cascade(_cascade).now();
                        } else {
                            saved = session.update(relatedBeans).cascade(_cascade).now();
                        }
                    } else {
                        RELATION relatedBean = (RELATION) relation.getGetManipulator().getValue(bean);
                        if (relatedBean!=null) {
                            relationClassField.getSetManipulator().setValue(relatedBean, beanPrimaryKeyValue);
                            if (_saveOrUpdate) {
                                saved = session.saveOrUpdate(relatedBean).cascade(_cascade).now();
                            } else {
                                saved = session.update(relatedBean).cascade(_cascade).now();
                            }
                        }
                    }
                    (( SetManipulator<BEAN, Object>) relation.getSetManipulator()).setValue(bean, saved);
                }
            }
        }
    }

    @Override
    public UpdateQuery<BEAN> setQueryTimeout(final int queryTimeout) {
        this.queryTimeout = queryTimeout;
        return this;
    }

    @Override
    public int getQueryTimeout() {
        return queryTimeout;
    }

    @Override
    public UpdateQuery<BEAN> values(final Map<String, Object> properties) {
        for (Entry<String, Object> propertyEntry : properties.entrySet()) {
            value(propertyEntry.getKey(), propertyEntry.getValue());
        }
        return this;
    }

    @Override
    public UpdateQuery<BEAN> cascade(final boolean cascade) {
        this._cascade = cascade;
        return this;
    }

    @Override
    public UpdateQuery<BEAN> saveOrUpdate(final boolean saveOrUpdate) {
        _saveOrUpdate = saveOrUpdate;
        return this;
    }

}
