package com.jpattern.orm.session;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import com.jpattern.orm.cache.CacheManager;
import com.jpattern.orm.dialect.DBProfile;
import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.mapper.IOrmClassTool;
import com.jpattern.orm.mapper.IOrmClassToolMap;
import com.jpattern.orm.query.QueryRoot;
import com.jpattern.orm.query.delete.DeleteQuery;
import com.jpattern.orm.query.delete.DeleteQueryOrm;
import com.jpattern.orm.query.delete.DeleteWhere;
import com.jpattern.orm.query.delete.orm.ADelete;
import com.jpattern.orm.query.delete.orm.Delete;
import com.jpattern.orm.query.find.CustomFindQuery;
import com.jpattern.orm.query.find.CustomFindQueryOrm;
import com.jpattern.orm.query.find.FindQuery;
import com.jpattern.orm.query.find.FindQueryOrm;
import com.jpattern.orm.query.find.FindWhere;
import com.jpattern.orm.query.find.orm.AFind;
import com.jpattern.orm.query.find.orm.Find;
import com.jpattern.orm.query.save.SaveQueryOrm;
import com.jpattern.orm.query.save.orm.ASave;
import com.jpattern.orm.query.save.orm.ASaveOrUpdate;
import com.jpattern.orm.query.save.orm.Save;
import com.jpattern.orm.query.save.orm.SaveOrUpdate;
import com.jpattern.orm.query.sql.PlainSqlExecutor;
import com.jpattern.orm.query.sql.SqlExecutor;
import com.jpattern.orm.query.update.CustomUpdateQuery;
import com.jpattern.orm.query.update.CustomUpdateQueryImpl;
import com.jpattern.orm.query.update.UpdateQueryOrm;
import com.jpattern.orm.query.update.orm.AUpdate;
import com.jpattern.orm.query.update.orm.Update;
import com.jpattern.orm.script.ScriptExecutor;
import com.jpattern.orm.script.ScriptExecutorImpl;
import com.jpattern.orm.transaction.OrmTransactionDefinition;
import com.jpattern.orm.transaction.Transaction;
import com.jpattern.orm.transaction.TransactionDefinition;
import com.jpattern.orm.validator.ValidatorService;

/**
 * 
 * @author Francesco Cina
 *
 * 27/giu/2011
 */
public class OrmSession implements SessionSqlPerformer {

    private final IOrmClassToolMap ormClassToolMap;
    private final SessionStrategy sessionStrategy;
    private final ValidatorService validatorService;
    private final DBProfile dbProfile;
    private final CacheManager cacheManager;

    public OrmSession(final IOrmClassToolMap ormClassToolMap, final SessionStrategy sessionStrategy, final ValidatorService validatorService, final CacheManager cacheManager, final DBProfile dbProfile) {
        this.ormClassToolMap = ormClassToolMap;
        this.sessionStrategy = sessionStrategy;
        this.validatorService = validatorService;
        this.cacheManager = cacheManager;
        this.dbProfile = dbProfile;
    }

    @Override
    public <BEAN> Delete<BEAN> delete(final BEAN bean) {
        return new ADelete<BEAN>() {
            @Override
            public int now() {
                Class<BEAN> clazz = (Class<BEAN>) bean.getClass();
                final IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                DeleteWhere<BEAN> query = deleteQuery(clazz).cascade(getCascade()).where();
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                Object[] pkValues = ormClassTool.getOrmPersistor().getPropertyValues(pks, bean);
                for (int i = 0; i < pks.length; i++) {
                    query.eq(pks[i], pkValues[i]);
                }
                return query.now();
            }
        };
    }

    @Override
    public final <BEAN> Delete<List<BEAN>> delete(final List<BEAN> beans) throws OrmException {
        return new ADelete<List<BEAN>>(){
            @Override
            public int now() {
                int result = 0;
                for (final BEAN bean : beans) {
                    result += delete(bean).cascade(getCascade()).now();
                }
                return result;
            }
        };
    }

    @Override
    public final <BEAN> DeleteQuery<BEAN> deleteQuery(final Class<BEAN> clazz) throws OrmException {
        final DeleteQueryOrm<BEAN> delete = new DeleteQueryOrm<BEAN>(clazz, ormClassToolMap, this);
        return delete;
    }

    @Override
    public <T> T doInTransaction(final TransactionCallback<T> transactionCallback)
            throws OrmException {
        return doInTransaction(new OrmTransactionDefinition(), transactionCallback);
    }

    @Override
    public <T> T doInTransaction(final TransactionDefinition transactionDefinition,
            final TransactionCallback<T> transactionCallback) throws OrmException {
        T result;
        Transaction tx = transaction(transactionDefinition);
        try {
            result = transactionCallback.doInTransaction(this);
            tx.commit();
        } catch (RuntimeException e) {
            tx.rollback();
            throw e;
        } catch (Error e) {
            tx.rollback();
            throw e;
        }
        return result;
    }

    @Override
    public final <BEAN> Find<BEAN> find(final BEAN bean) throws OrmException {
        IOrmClassTool<BEAN> ormClassTool = (IOrmClassTool<BEAN>) getOrmClassToolMap().getOrmClassTool(bean.getClass());
        String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
        Object[] values =  ormClassTool.getOrmPersistor().getPropertyValues(pks, bean);
        return find((Class<BEAN>) bean.getClass(), values);
    }

    @Override
    public final <BEAN> Find<BEAN> find(final Class<BEAN> clazz, final Object value) throws OrmException {
        return this.find(clazz, new Object[]{value});
    }

    @Override
    public final <BEAN> Find<BEAN> find(final Class<BEAN> clazz, final Object[] values) throws OrmException {
        return new AFind<BEAN>() {
            @Override
            public BEAN get() {
                IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                FindWhere<BEAN> query = findQuery(clazz, clazz.getSimpleName() + QueryRoot.ALIAS_FIXED_SUFFIX)
                        .lazy(isLazy()).cache(getCache()).ignore(getIgnoredFields()).where();
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                for (int i = 0; i < pks.length; i++) {
                    query.eq(pks[i], values[i]);
                }
                return query.maxRows(1).get();
            }

            @Override
            public BEAN getUnique() {
                IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                FindWhere<BEAN> query = findQuery(clazz, clazz.getSimpleName() + QueryRoot.ALIAS_FIXED_SUFFIX)
                        .lazy(isLazy()).cache(getCache()).ignore(getIgnoredFields()).where();
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                for (int i = 0; i < pks.length; i++) {
                    query.eq(pks[i], values[i]);
                }
                return query.maxRows(1).getUnique();
            }

            @Override
            public boolean exist() {
                IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                FindWhere<BEAN> query = findQuery(clazz).where();
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                for (int i = 0; i < pks.length; i++) {
                    query.eq(pks[i], values[i]);
                }
                return query.maxRows(1).getRowCount()>0;
            }
        };
    }

    @Override
    public final <BEAN> FindQuery<BEAN> findQuery(final Class<BEAN> clazz) throws OrmException {
        return findQuery(clazz, clazz.getSimpleName());
    }

    @Override
    public final <BEAN> FindQuery<BEAN> findQuery(final Class<BEAN> clazz, final String alias) throws OrmException {
        final FindQueryOrm<BEAN> query = new FindQueryOrm<BEAN>(ormClassToolMap , this, clazz, cacheManager, alias);
        return query;
    }

    @Override
    public final CustomFindQuery findQuery(final String selectFields, final Class<?> clazz, final String alias ) throws OrmException {
        return findQuery(new String[]{selectFields}, clazz, alias);
    }

    @Override
    public final CustomFindQuery findQuery(final String[] selectFields, final Class<?> clazz, final String alias ) throws OrmException {
        final CustomFindQueryOrm query = new CustomFindQueryOrm(selectFields, ormClassToolMap , this, clazz, alias);
        return query;
    }

    public final IOrmClassToolMap getOrmClassToolMap() {
        return ormClassToolMap;
    }

    @Override
    public <BEAN> Save<BEAN> save(final BEAN bean) {
        return new ASave<BEAN>() {
            @Override
            public BEAN now() {
                validatorService.validator(bean).validateThrowException();
                Class<BEAN> clazz = (Class<BEAN>) bean.getClass();
                final IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                BEAN newBean = ormClassTool.getOrmPersistor().clone(bean);
                return new SaveQueryOrm<BEAN>(newBean, ormClassToolMap, OrmSession.this, dbProfile)
                        .cascade(getCascade()).perform();
            }
        };
    }

    @Override
    public <BEAN> Save<List<BEAN>> save(final Collection<BEAN> beans) throws OrmException {
        return new ASave<List<BEAN>>() {
            @Override
            public List<BEAN> now() {
                final List<BEAN> result = new ArrayList<BEAN>();
                for (final BEAN bean : beans) {
                    result.add(save(bean).cascade(getCascade()).now());
                }
                return result;
            }
        };
    }

    @Override
    public <BEAN> SaveOrUpdate<BEAN> saveOrUpdate(final BEAN bean) throws OrmException {
        return new ASaveOrUpdate<BEAN>() {
            @Override
            public BEAN now() {
                validatorService.validator(bean).validateThrowException();
                Class<BEAN> clazz = (Class<BEAN>) bean.getClass();
                final IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                BEAN newBean = ormClassTool.getOrmPersistor().clone(bean);

                if (ormClassTool.getOrmPersistor().hasGenerator()) {
                    if(ormClassTool.getOrmPersistor().useGenerators(bean)) {
                        return new SaveQueryOrm<BEAN>(newBean, ormClassToolMap, OrmSession.this, dbProfile)
                                .cascade(getCascade()).saveOrUpdate(true).perform();
                    }
                    return new UpdateQueryOrm<BEAN>(newBean, ormClassToolMap, OrmSession.this)
                            .cascade(getCascade()).saveOrUpdate(true).perform();

                }
                if (find(bean).exist()) {
                    return new UpdateQueryOrm<BEAN>(newBean, ormClassToolMap, OrmSession.this)
                            .cascade(getCascade()).saveOrUpdate(true).perform();
                }
                return new SaveQueryOrm<BEAN>(newBean, ormClassToolMap, OrmSession.this, dbProfile)
                        .cascade(getCascade()).saveOrUpdate(true).perform();
            }
        };
    }

    @Override
    public <BEAN> SaveOrUpdate<List<BEAN>> saveOrUpdate(final Collection<BEAN> beans) throws OrmException {
        return new ASaveOrUpdate<List<BEAN>>() {
            @Override
            public List<BEAN> now() {
                final List<BEAN> result = new ArrayList<BEAN>();
                for (final BEAN bean : beans) {
                    result.add(saveOrUpdate(bean).cascade(getCascade()).now());
                }
                return result;
            }
        };
    }

    @Override
    public final ScriptExecutor scriptExecutor() throws OrmException {
        return new ScriptExecutorImpl(this);
    }

    @Override
    public final SqlExecutor sqlExecutor() {
        return new PlainSqlExecutor(this);
    }

    @Override
    public SqlPerformer sqlPerformer() throws OrmException {
        return new PlainSqlPerformer(sessionStrategy.sqlPerformerStrategy(), ormClassToolMap.getTypeFactory());
    }

    @Override
    public final Transaction transaction() throws OrmException {
        return this.transaction(new OrmTransactionDefinition());
    }

    @Override
    public Transaction transaction(
            final TransactionDefinition transactionDefinition) throws OrmException {
        return sessionStrategy.getTransaction(transactionDefinition);
    }

    @Override
    public <BEAN> Update<BEAN> update(final BEAN bean) throws OrmException {
        return new AUpdate<BEAN>() {
            @Override
            public BEAN now() {
                validatorService.validator(bean).validateThrowException();
                Class<BEAN> clazz = (Class<BEAN>) bean.getClass();
                final IOrmClassTool<BEAN> ormClassTool = getOrmClassToolMap().getOrmClassTool(clazz);
                BEAN newBean = ormClassTool.getOrmPersistor().clone(bean);
                return new UpdateQueryOrm<BEAN>(newBean, ormClassToolMap, OrmSession.this)
                        .cascade(getCascade()).perform();
            }
        };
    }

    @Override
    public <BEAN> Update<List<BEAN>> update(final Collection<BEAN> beans) throws OrmException {
        return new AUpdate<List<BEAN>>(){
            @Override
            public List<BEAN> now() {
                final List<BEAN> result = new ArrayList<BEAN>();
                for (final BEAN bean : beans) {
                    result.add(update(bean).cascade(getCascade()).now());
                }
                return result;
            }
        };
    }

    @Override
    public final <BEAN> CustomUpdateQuery updateQuery(final Class<BEAN> clazz) throws OrmException {
        final CustomUpdateQueryImpl update = new CustomUpdateQueryImpl(clazz, ormClassToolMap, this);
        return update;
    }

}