package com.jpattern.orm.session;

import java.util.Collection;
import java.util.List;

import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.query.delete.DeleteQuery;
import com.jpattern.orm.query.delete.orm.Delete;
import com.jpattern.orm.query.find.CustomFindQuery;
import com.jpattern.orm.query.find.FindQuery;
import com.jpattern.orm.query.find.orm.Find;
import com.jpattern.orm.query.save.orm.Save;
import com.jpattern.orm.query.save.orm.SaveOrUpdate;
import com.jpattern.orm.query.sql.SqlExecutor;
import com.jpattern.orm.query.update.CustomUpdateQuery;
import com.jpattern.orm.query.update.orm.Update;
import com.jpattern.orm.script.ScriptExecutor;
import com.jpattern.orm.transaction.Transaction;
import com.jpattern.orm.transaction.TransactionDefinition;

/**
 * 
 * @author Francesco Cina
 *
 * 21/mag/2011
 * 
 */
public interface Session {

    /**
     * Delete one bean from the database
     * 
     * @param bean
     * @param cascade
     * @return
     */
    <BEAN> Delete<BEAN> delete(BEAN bean) throws OrmException;

    /**
     * Delete the beans from the database
     * @param <BEAN>
     * @param beans the beans to delete
     * @throws OrmException
     * @return
     */
    <BEAN> Delete<List<BEAN>> delete(List<BEAN> beans) throws OrmException;

    /**
     * Delete entries from a specific table
     * @param clazz the TABLE related Class
     * @throws OrmException
     */
    <BEAN> DeleteQuery<BEAN> deleteQuery(Class<BEAN> clazz) throws OrmException;

    /**
     * Execute a block of code inside a Transaction or participate to an existing one
     * @param transactionCallback
     * @return
     */
    <T> T doInTransaction(TransactionCallback<T> transactionCallback);

    /**
     * Execute a block of code inside a Transaction or participate to an existing one
     * @param transactionCallback
     * @return
     */
    <T> T doInTransaction(TransactionDefinition transactionDefinition, TransactionCallback<T> transactionCallback);

    /**
     * Find a bean using the bean type and id(s).
     * @param <BEAN>
     * @param bean
     * @return
     * @throws OrmException
     */
    <BEAN> Find<BEAN> find(BEAN bean) throws OrmException;

    /**
     * Find a bean using its ID.
     * 
     * @param <BEAN>
     * @param clazz The Class of the bean to load
     * @param idValue the value of the identifying column of the bean
     * @return
     */
    <BEAN> Find<BEAN> find(Class<BEAN> clazz, Object idValue);

    /**
     * Find a bean using its IDs.
     * 
     * @param <BEAN>
     * @param clazz The Class of the bean to load
     * @param idValues an ordered array with the values of the identifying columns of the bean
     * @return
     * @throws OrmException
     */
    <BEAN> Find<BEAN> find(Class<BEAN> clazz, Object[] idValues) throws OrmException;

    /**
     * Create a new query
     * @param <BEAN>
     * @param clazz The class of the bean that will be retrieved by the query execution. The simple class name will be used as alias for the class
     * @return
     * @throws OrmException
     */
    <BEAN> FindQuery<BEAN> findQuery(Class<BEAN> clazz) throws OrmException;

    /**
     * Create a new query
     * @param <BEAN>
     * @param clazz The class of the bean that will be retrieved by the query execution.
     * @param alias The alias of the class in the query.
     * @return
     * @throws OrmException
     */
    <BEAN> FindQuery<BEAN> findQuery(Class<BEAN> clazz, String alias) throws OrmException;

    /**
     * Create a new query
     * @param selectFields the comma separated fields to fetch
     * @param clazz The class of the object that will be retrieved by the query execution.
     * @param alias The alias of the class in the query.
     * @return
     * @throws OrmException
     */
    CustomFindQuery findQuery(String selectFields, Class<?> clazz, String alias ) throws OrmException;

    /**
     * Create a new query
     * @param selectFields the name of the fields to fetch
     * @param clazz The class of the object that will be retrieved by the query execution.
     * @param alias The alias of the class in the query.
     * @return
     * @throws OrmException
     */
    CustomFindQuery findQuery(String[] selectFields, Class<?> clazz, String alias ) throws OrmException;

    /**
     * Persist the new bean in the database
     * @param <BEAN>
     * @param bean
     * @throws OrmException
     * @return
     */
    <BEAN> Save<BEAN> save(BEAN bean);

    /**
     * Persist the new beans in the database
     * @param beans the beans to persist
     * @param cascade whether to persist the children recursively
     * @return
     * @throws OrmException
     */
    <BEAN> Save<List<BEAN>> save(Collection<BEAN> beans) throws OrmException;

    /**
     * For each bean in the list, update the bean if it exists,
     * otherwise saves it
     * @param bean the bean to persist
     * @return
     * @throws OrmException
     */
    <BEAN> SaveOrUpdate<BEAN> saveOrUpdate(BEAN bean) throws OrmException;

    /**
     * For each bean in the list, update the bean if it exists,
     * otherwise saves it
     * @param beans the beans to persist
     * @param cascade whether to saveOrUpdate the children recursively
     * @return
     * @throws OrmException
     */
    <BEAN> SaveOrUpdate<List<BEAN>> saveOrUpdate(Collection<BEAN> beans) throws OrmException;

    /**
     * A script executor useful to execute multiple sql statement from files.
     * @return
     * @throws OrmException
     */
    ScriptExecutor scriptExecutor() throws OrmException;

    /**
     * An executor to perform any kind of plain SQL statements.
     * @return
     */
    SqlExecutor sqlExecutor();

    /**
     * Begin a transaction or participate to an existing one using the default ITransactionDefinition
     * @return
     * @throws OrmException
     */
    Transaction transaction() throws OrmException;

    /**
     * Begin a or participate to an existing one depending on the specific transactionDefinition
     * @return
     * @throws OrmException
     */
    Transaction transaction(TransactionDefinition transactionDefinition) throws OrmException;

    /**
     * @param aggregatedUser
     * @return
     */
    <BEAN> Update<BEAN> update(BEAN bean) throws OrmException;

    /**
     * Update the values of the existing beans in the database
     * @param <BEAN>
     * @param beans the beans to update
     * @throws OrmException
     * @return
     */
    <BEAN> Update<List<BEAN>> update(Collection<BEAN> beans) throws OrmException;

    /**
     * Update the entries of a specific TABLE
     * @param clazz the TABLE related Class
     * @throws OrmException
     */
    <BEAN> CustomUpdateQuery updateQuery(Class<BEAN> clazz) throws OrmException;

}
