package com.jpattern.orm.session.datasource;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Savepoint;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jpattern.orm.exception.OrmException;
import com.jpattern.orm.exception.OrmRollbackException;

public class DataSourceConnection implements IConnection {

    private final Logger logger = LoggerFactory.getLogger(this.getClass());
    private ConnectionWrapper connectionWrapper;
    private List<IConnectionCaller> connectionCallers = new ArrayList<IConnectionCaller>();
    private boolean rollbackOnly = false;
    private boolean readOnly = false;
    private boolean valid = true;

    public DataSourceConnection(final DataSource dataSource, final boolean readOnly) {
        connectionWrapper = new ConnectionWrapper(dataSource);
        setReadOnly(readOnly);
    }

    @Override
    public void setTransactionIsolation(final int transactionIsolation) throws OrmException {
        try {
            connectionWrapper.setTransactionIsolation(transactionIsolation);
        } catch (SQLException e) {
            throw new OrmException(e);
        }
    }

    @Override
    public boolean isClosed() throws OrmException {
        try {
            return connectionWrapper.isClosed();
        } catch (SQLException e) {
            throw new OrmException(e);
        }
    }

    @Override
    public void rollback() throws OrmException {
        if (getLogger().isDebugEnabled()) {
            getLogger().debug("rollback called"); //$NON-NLS-1$
        }
        if ((connectionCallers.size()==1) && !isReadOnly()) {
            try {
                if (getLogger().isDebugEnabled()) {
                    getLogger().debug("performing rollback"); //$NON-NLS-1$
                }
                connectionWrapper.rollback();
            } catch (SQLException e) {
                throw new OrmException(e);
            }
        }
    }

    @Override
    public void commit() throws OrmException {
        if (getLogger().isDebugEnabled()) {
            getLogger().debug("commit called"); //$NON-NLS-1$
        }
        if ((connectionCallers.size()==1) && !isReadOnly()) {
            if (rollbackOnly) {
                rollback();
                throw new OrmRollbackException("Transaction rolled back because it has been marked as rollback-only"); //$NON-NLS-1$
            }
            try {
                if (getLogger().isDebugEnabled()) {
                    getLogger().debug("performing commit"); //$NON-NLS-1$
                }
                connectionWrapper.commit();
            } catch (SQLException e) {
                throw new OrmException(e);
            }
        }
    }

    @Override
    public PreparedStatement prepareStatement(final String sql) throws OrmException {
        try {
            return connectionWrapper.prepareStatement(sql);
        } catch (SQLException e) {
            throw new OrmException(e);
        }
    }

    @Override
    public PreparedStatement prepareStatement(final String sql, final String[] generatedColumnNames) throws OrmException {
        try {
            return connectionWrapper.prepareStatement(sql, generatedColumnNames) ;
        } catch (SQLException e) {
            throw new OrmException(e);
        }
    }

    @Override
    public IStatement createStatement() throws OrmException {
        try {
            return new StatementWrapper(connectionWrapper.createStatement() , this);
        } catch (SQLException e) {
            throw new OrmException(e);
        }
    }

    @Override
    public void addCaller(final IConnectionCaller connectionCaller) throws OrmException {
        if (getLogger().isDebugEnabled()) {
            getLogger().debug("Caller " + connectionCaller + " registered"); //$NON-NLS-1$ //$NON-NLS-2$
        }
        connectionCallers.add(connectionCaller);
    }

    @Override
    public void close(final IConnectionCaller connectionCaller) throws OrmException {
        if (getLogger().isDebugEnabled()) {
            getLogger().debug("Caller " + connectionCaller + " called a close on transaction"); //$NON-NLS-1$ //$NON-NLS-2$
        }
        connectionCallers.remove(connectionCaller);
        if (connectionCallers.isEmpty()) {
            try {
                if (getLogger().isDebugEnabled()) {
                    getLogger().debug("Closing SQL connection"); //$NON-NLS-1$
                }
                connectionWrapper.close();
                valid = false;
            } catch (SQLException e) {
                throw new OrmException(e);
            }
        }
    }

    @Override
    public void setRollbackOnly() throws OrmException {
        if (getLogger().isDebugEnabled()) {
            getLogger().debug("Transaction sets as a rollbackOnly"); //$NON-NLS-1$
        }
        rollbackOnly = true;
    }

    @Override
    public void setReadOnly(final boolean readOnly) throws OrmException {
        this.readOnly = readOnly;
    }

    @Override
    public boolean isValid() throws OrmException {
        return valid;
    }

    public boolean isReadOnly() {
        return readOnly;
    }

    protected Logger getLogger() {
        return logger;
    }

    class ConnectionWrapper {

        private java.sql.Connection connection;
        private final DataSource dataSource;
        private Savepoint savepoint;

        ConnectionWrapper(final DataSource dataSource) {
            this.dataSource = dataSource;
        }

        public void setTransactionIsolation(final int transactionIsolation) throws SQLException {
            validateConnection();
            connection.setTransactionIsolation(transactionIsolation);
        }

        public void rollback() throws SQLException {
            if (connection!=null) {
                if (savepoint==null) {
                    connection.rollback();
                } else {
                    connection.rollback(savepoint);
                }

            }
        }

        public void rollback(final Savepoint rollbackToSavepoint) throws SQLException {
            if (connection!=null) {
                connection.rollback(rollbackToSavepoint);
            }
        }

        public void commit() throws SQLException {
            if (connection!=null) {
                connection.commit();
            }
        }

        public Statement createStatement() throws SQLException {
            validateConnection();
            return connection.createStatement();
        }

        public PreparedStatement prepareStatement(final String sql) throws SQLException {
            validateConnection();
            return connection.prepareStatement(sql);
        }

        public PreparedStatement prepareStatement(final String sql, final String[] generatedColumnNames) throws SQLException {
            validateConnection();
            return connection.prepareStatement(sql, generatedColumnNames);
        }

        public void close() throws SQLException {
            if (connection!=null) {
                connection.close();
            }
        }

        public boolean isClosed() throws SQLException {
            if (connection!=null) {
                return connection.isClosed();
            }
            return true;
        }

        private void validateConnection() throws SQLException {
            if (!isValid()) {
                throw new OrmException("Not possible to open a new java.sql.Connection"); //$NON-NLS-1$
            }
            if (isClosed()) {
                if (getLogger().isDebugEnabled()) {
                    getLogger().debug("Sql connection is null or closed. Opening a new sql connection."); //$NON-NLS-1$
                }
                Connection sqlConn = dataSource.getConnection();
                if (!isReadOnly()) {
                    savepoint = sqlConn.setSavepoint();
                }
                connection = sqlConn;
            }
        }

    }
}
